#ifndef _SPARC_SYSEN_H
#define _SPARC_SYSEN_H

/* sysen.h:  Bit fields within the "System Enable" register accessed via
             the ASI_CONTROL address space at address AC_SYSENABLE.

   Copyright (C) 1994 David S. Miller (davem@caip.rutgers.edu)
*/

#define SENABLE_DVMA   0x20   /* enable dvma transfers */
#define SENABLE_CACHE  0x10   /* enable VAC cache */
#define SENABLE_RESET  0x04   /* reset the whole machine, danger Will Robinson */

#endif /* _SPARC_SYSEN_H */
